/********************************************************
 ** unpic  - SMS2/QDOS to TIFF, GIF (c) Jonathan Hudson 1997
 **
 ********************************************************/

/*****************************************************************
 * Portions of this code Copyright (C) 1989 by Michael Mauldin.
 * Permission is granted to use this file in whole or in
 * part for any purpose, educational, recreational or commercial,
 * provided that this copyright notice is retained unchanged.
 * This software is available to all free of charge by anonymous
 * FTP and in the UUNET archives.
 *
 *
 * Authors:  Michael Mauldin (mlm@cs.cmu.edu)
 *           David Rowley (mgardi@watdcsu.waterloo.edu)
 *
 * Based on: compress.c - File compression ala IEEE Computer, June 1984.
 *
 *	Spencer W. Thomas       (decvax!harpo!utah-cs!utah-gr!thomas)
 *	Jim McKie               (decvax!mcvax!jim)
 *	Steve Davies            (decvax!vax135!petsd!peora!srd)
 *	Ken Turkowski           (decvax!decwrl!turtlevax!ken)
 *	James A. Woods          (decvax!ihnp4!ames!jaw)
 *	Joe Orost               (decvax!vax135!petsd!joe)
 *****************************************************************/

/*======================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>

#define VERSION "0.06"

#ifdef QDOS
char _prog_name[] = "unpic";
# include <memory.h>
typedef unsigned short ushort;
typedef unsigned long ulong;
typedef unsigned int uint;
typedef unsigned short u_short;
typedef unsigned long u_long;
# define OSNAME "SMS/QDOS"
#elif defined(__linux__)
# define OSNAME "Linux"
#else
# define OSNAME "????"
#endif

#ifndef NOTIF
#include "tiffio.h"
#else
typedef unsigned short ushort;
typedef unsigned long ulong;
typedef unsigned int uint;
typedef unsigned short u_short;
typedef unsigned long u_long;
#endif
typedef unsigned char uchar;
static short verbose = 0;

#define PARM(p1__) p1__

#ifndef LZW_CLEAN
static void fputw PARM ((short, FILE *));
static void gcompress PARM ((int, FILE *, FILE *, int, int, int,
			     void (*f) (uchar *, uchar *, short, short)));
static void output PARM ((int));
static void cl_block PARM ((void));
static void cl_hash PARM ((long));
static void char_init PARM ((void));
static void char_out PARM ((int));
static void flush_char PARM ((void));
#define METHODS "X:x:PCX\000F:tn:TIFF (no compress)\000I:tp:TIFF (packbits)\000T:t:TIFF (LZW)\000G:g:GIF\000S:s:EPSF\000P:p:PNG\000\000\n"
#else
#define METHODS "X:x:PCX\000S:s:EPSF\000P:p:PNG\000\000\n"
#endif

#ifdef __i386__
# define VAX_ENDIAN 1
#endif

#define swapword(v) ((ushort)(v) >> 8) + ((ushort)(v) << 8)

#ifdef VAX_ENDIAN
# define VAXSHORT(v) v
# define VAXLONG(v) v
#else
# define VAXSHORT(v) (swapword(v))
# define VAXLONG(v) ((ulong)swapword(v >> 16) | (swapword ((v) & 0xffff) << 16))
#endif

typedef struct
{
    unsigned short flag;
    short xpix;
    short ypix;
    short llen;
    uchar mode[2];
}
QHEAD;

typedef struct
{
    FILE *inp;
    FILE *fp;
    QHEAD qh;
    short flag;
    short qlar;
}
QLS;

typedef struct
{
    uchar manuf;		/* always 0x0a */
    uchar version;		/* version info = 0,2,3,5 */
    uchar encoding;		/* 1=RLE */
    uchar bpp;			/* bits per pixel per plane */
    ushort x1;			/* picture dimensions */
    ushort y1;
    ushort x2;
    ushort y2;
    ushort hres;		/* horz. resolution */
    ushort vres;		/* vert. resolution */
    uchar palette[16 * 3];	/* color palette */
    uchar reserved;
    uchar nplanes;		/* number of color planes */
    ushort bpl;			/* number of bytes per line (uncompressed) */
    ushort palinfo;		/* palette info 1=color, 2=grey */
    uchar xtra[58];		/* fill out header to 128 bytes */
}
pcx_header;


void *xmalloc(size_t n)
{
    void *p;
    if((p = malloc(n)) != NULL)
    {
	memset(p, '\0', n);
    }
    else
    {
	exit(0);
    }
    return p;
}

static void MakeChunkyPix4 (uchar * cpix, uchar * qpix, short xpix, short s)
{
    uchar *q, *p;
    ushort k;
    ushort j;
    ushort l;
    ushort gb, rb;
    int llen, m;
    short ls;

    p = cpix;
    q = qpix;

    llen = ((xpix + 3) & ~3);
    memset (cpix, '\0', llen);

    k = 7;
    l = 0;

    ls = (s) ? 8 - s : 0;

    for (m = 0; m < xpix; m++)
    {
	j = 1 << k;
	gb = (*q & j);
	if (gb)
	{
	    *p |= (2 << ls);
	}

	rb = *(q + 1) & j;
	if (rb)
	{
	    *p |= (1 << ls);
	}

	if (s)
	{
	    ls -= s;
	    if (ls < 0)
	    {
		ls = 8 - s;
		p++;
	    }
	}
	else
	    p++;

	if (k == 0)
	{
	    k = 7;
	    q += 2;
	}
	else
	{
	    k--;
	}
    }
}
static void MakeChunkyPix2 (uchar * cpix, uchar * qpix, short xpix, short s)
{
    uchar *q, *p;
    ushort k;
    ushort j;
    ushort l;
    ushort gb;
    int llen, m;
    short ls = 0;

    p = cpix;
    q = qpix;

    llen = ((xpix + 3) & ~3);

    if (s)
	llen >>= 3;
    memset (cpix, '\0', llen);

    k = 7;
    l = 0;

    for (m = 0; m < xpix; m++)
    {
	j = 1 << k;
	gb = (*q & j);

	if (s)
	    ls = k;

	if (gb)
	{
	    *p |= (1 << ls);
	}

	if (ls == 0)
	{
	    p++;
	}

	if (k == 0)
	{
	    k = 7;
	    q++;
	}
	else
	{
	    k--;
	}
    }
}

static void MakeChunkyPix8 (uchar * cpix, uchar * qpix, short xpix, short s)
{
    uchar *q, *p;
    ushort k;
    ushort j;
    ushort l;
    ushort gb, rb, bb;
    int llen, m;
    short ls;
    llen = ((xpix + 3) & ~3);

    p = cpix;
    q = qpix;

    memset (cpix, '\0', llen);

    k = 7;
    l = 0;

    ls = (s) ? 8 : 0;

    for (m = 0; m < xpix; m++)
    {
	short v;
	j = 1 << k;
	gb = (*q & j);
	rb = (*(q + 1)) & j;
	k--;
	j = 1 << k;
	bb = (*(q + 1)) & j;
	v = 0;
	if (rb)
	{
	    v = 1;
	}
	if (gb)
	{
	    v |= 2;
	}
	if (bb)
	{
	    v |= 4;
	}

	if (s)
	{
	    ls -= s;
	    *p |= (v << ls);
	    ls -= s;
	    *p |= (v << ls);
	    if (ls == 0)
	    {
		ls = 8;
		p++;
	    }
	}
	else
	{
	    *p++ = v;
	    *p = *(p - 1);
	    p++;
	}

	if (k == 0)
	{
	    k = 7;
	    q += 2;
	}
	else
	{
	    k--;
	}
    }
}
void usage (void)
{
    fputs ("unpic -gnptGNPv -mN -TC  infile outfile\n", stderr);
    exit (0);
}

void ObnoxiousMsg (void)
{
    fprintf (stderr, "The Graphics Interchange Format(c) is\n");
    fprintf (stderr, "the Copyright Property of CompuServe Incorporated.\n");
    fprintf (stderr, "GIF(sm) is a Service Mark property of CompuServe Incorporated.\n");
}


#if !defined (NOTIF) && !defined(LZW_CLEAN)
#define	WIDEN(p1) ((((long)p1)*((1L<<16)-1))/255)

static void Colour16 (TIFF * tif, uchar * rmap,
		      uchar * gmap, uchar * bmap, int ns, int nc)
{
    ushort *red, *green, *blue;
    int i, j;

    j = 1 << ns;

    red = xmalloc (j * 3 * sizeof (short));
    green = red + j;
    blue = red + j + j;

    /* convert 8-bit colormap to 16-bit */
    for (i = 0; i < nc; i++)
    {
	red[i] = WIDEN (rmap[i]);
	green[i] = WIDEN (gmap[i]);
	blue[i] = WIDEN (bmap[i]);
    }
    TIFFSetField (tif, TIFFTAG_COLORMAP, red, green, blue);
    free (red);
}

static int WriteTIFF (char *fnam, FILE * inp, int w, int h, int llen,
		      uchar * rmap, uchar * gmap, uchar * bmap,
		      int ncol, int comp, char *comment,
		      void (*func) (uchar *, uchar *, short, short))
{
    TIFF *tif;
    uchar *data, *qldat;
    int hh;
    short ns;

    data = xmalloc (llen * 8);
    qldat = xmalloc (llen);

    tif = TIFFOpen (fnam, "w");

    TIFFSetField (tif, TIFFTAG_IMAGEWIDTH, w);
    TIFFSetField (tif, TIFFTAG_IMAGELENGTH, h);
    TIFFSetField (tif, TIFFTAG_COMPRESSION, comp);

    if (comment && strlen (comment))
    {
	TIFFSetField (tif, TIFFTAG_IMAGEDESCRIPTION, comment);
    }

    ns = (ncol == 2) ? 1 : (ncol == 4) ? 2 : 4;

    TIFFSetField (tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (tif, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (tif, TIFFTAG_ROWSPERSTRIP, h);
    TIFFSetField (tif, TIFFTAG_BITSPERSAMPLE, ns);
    TIFFSetField (tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
    TIFFSetField (tif, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);

    Colour16 (tif, rmap, gmap, bmap, ns, ncol);

    for (hh = 0; hh < h; )
    {
	fread (qldat, 1, llen, inp);
	(func) (data, qldat, w, ns);
	TIFFWriteScanline (tif, data, hh, 0);
	hh++;
	if (verbose)
	{
	    fprintf (stderr, "%d/%d%c", hh, h, verbose);
	    fflush (stderr);
	}
    }
    TIFFClose (tif);
    free(qldat);
    free(data);
    return 0;
}
#endif

#ifndef LZW_CLEAN
int WriteGIF (FILE * fp, FILE * ifp, int w, int h, int llen,
	      uchar * rmap, uchar * gmap, uchar * bmap,
	      int nc, char *comment,
	      void (*func) (uchar *, uchar *, short, short))
{
    int wid;
    int ColorMapSize, InitCodeSize, BitsPerPixel;
    int i;

    for (i = 1; i < 8; i++)
	if ((1 << i) >= nc)
	    break;

    BitsPerPixel = i;

    ColorMapSize = 1 << BitsPerPixel;

    wid = ((w + 3) & ~3);

    if (BitsPerPixel <= 1)
	InitCodeSize = 2;
    else
	InitCodeSize = BitsPerPixel;

    if (!fp)
    {
	fprintf (stderr, "WriteGIF: file not open for writing\n");
	return (1);
    }

    if (comment && strlen (comment) > (size_t) 0)
	fwrite ("GIF89a", (size_t) 1, (size_t) 6, fp);
    else
	fwrite ("GIF87a", (size_t) 1, (size_t) 6, fp);

    fputw (wid, fp);		/* screen descriptor */
    fputw (h, fp);

    i = 0x80;			/* Yes, there is a color map */
    i |= (2 - 1) << 4;		/* OR in the color resolution (hardwired 8) */
    i |= (BitsPerPixel - 1);	/* OR in the # of bits per pixel */

    fputc (i, fp);
    fputc (0, fp);		/* background color */
    fputc (0, fp);		/* future expansion byte */

    for (i = 0; i < ColorMapSize; i++)
    {				/* write out Global colormap */
	fputc (rmap[i], fp);
	fputc (gmap[i], fp);
	fputc (bmap[i], fp);
    }

    if (comment && strlen (comment) > (size_t) 0)
    {				/* write comment blocks */
	char *sp;
	int i, blen;

	fputc (0x21, fp);	/* EXTENSION block */
	fputc (0xFE, fp);	/* comment extension */

	sp = comment;
	while ((blen = strlen (sp)) > 0)
	{
	    if (blen > 255)
		blen = 255;
	    fputc (blen, fp);
	    for (i = 0; i < blen; i++, sp++)
		fputc (*sp, fp);
	}
	fputc (0, fp);		/* zero-length data subblock to end extension */
    }

    fputc (',', fp);		/* image separator */

    /* Write the Image header */
    fputw (0, fp);
    fputw (0, fp);
    fputw (wid, fp);
    fputw (h, fp);
    fputc (0x00, fp);
    fputc (InitCodeSize, fp);

    gcompress (InitCodeSize + 1, fp, ifp, wid, h, llen, func);

    fputc (0, fp);		/* Write out a Zero-length packet (EOF) */
    fputc (';', fp);		/* Write GIF file terminator */

    if (ferror (fp))
	return -1;
    return (0);
}

static void fputw (short i, FILE * fp)
{
    unsigned short c;
    c = (ushort) i & 0xff;
    putc (c, fp);
    c = ((ushort) i >> 8);
    putc (c, fp);
}


/***********************************************************************/

static unsigned long cur_accum = 0;
static int cur_bits = 0;

#define min(a,b)        ((a>b) ? b : a)

#define BITS	12
#define HSIZE  5003		/* 80% occupancy */

typedef unsigned char char_type;


static int n_bits;		/* number of bits/code */
static int maxbits = BITS;	/* user settable max # bits/code */
static int maxcode;		/* maximum code, given n_bits */
static int maxmaxcode = 1 << BITS;	/* NEVER generate this */

#define MAXCODE(n_bits)     ( (1 << (n_bits)) - 1)

static long htab[HSIZE];
static unsigned short codetab[HSIZE];
#define HashTabOf(i)   htab[i]
#define CodeTabOf(i)   codetab[i]

static int hsize = HSIZE;	/* for dynamic table sizing */

/*
 * To save much memory, we overlay the table used by compress() with those
 * used by decompress().  The tab_prefix table is the same size and type
 * as the codetab.  The tab_suffix table needs 2**BITS characters.  We
 * get this from the beginning of htab.  The output stack uses the rest
 * of htab, and contains characters.  There is plenty of room for any
 * possible stack (stack used to be 8000 characters).
 */

#define tab_prefixof(i) CodeTabOf(i)
#define tab_suffixof(i)        ((char_type *)(htab))[i]
#define de_stack               ((char_type *)&tab_suffixof(1<<XV_BITS))

static int free_ent = 0;	/* first unused entry */

/*
 * block compression parameters -- after all codes are used up,
 * and compression rate changes, start over.
 */
static int clear_flg = 0;

static long int in_count = 1;	/* length of input */
static long int out_count = 0;	/* # of codes output (for debugging) */

/*
 * compress stdin to stdout
 *
 * Algorithm:  use open addressing double hashing (no chaining) on the 
 * prefix code / next character combination.  We do a variant of Knuth's
 * algorithm D (vol. 3, sec. 6.4) along with G. Knott's relatively-prime
 * secondary probe.  Here, the modular division first probe is gives way
 * to a faster exclusive-or manipulation.  Also do block compression with
 * an adaptive reset, whereby the code table is cleared when the compression
 * ratio decreases, but after the table fills.  The variable-length output
 * codes are re-sized at this point, and a special CLEAR code is generated
 * for the decompressor.  Late addition:  construct the table according to
 * file size for noticeable speed improvement on small files.  Please direct
 * questions about this implementation to ames!jaw.
 */

static int g_init_bits;
static FILE *g_outfile;

static int ClearCode;
static int EOFCode;


/********************************************************/
static void gcompress (int init_bits, FILE * outfile, FILE * inp,
		       int w, int h, int llen,
		       void (*cvt) (uchar *, uchar *, short, short))
{
    long fcode;
    int i = 0;
    int c;
    int ent = 0;
    int disp;
    int hsize_reg;
    int hshift;
    uchar *row, *data, *qldat;
    int hh;

    short first = 1;
    data = xmalloc (llen * 8);
    qldat = xmalloc (llen);

    /*
     * Set up the globals:  g_init_bits - initial number of bits
     *                      g_outfile   - pointer to output file
     */
    g_init_bits = init_bits;
    g_outfile = outfile;

    /* initialize 'compress' globals */
    maxbits = BITS;
    maxmaxcode = 1 << BITS;
    bzero ((char *) htab, sizeof (htab));
    bzero ((char *) codetab, sizeof (codetab));
    hsize = HSIZE;
    free_ent = 0;
    clear_flg = 0;
    in_count = 1;
    out_count = 0;
    cur_accum = 0;
    cur_bits = 0;

    /*
     * Set up the necessary values
     */
    out_count = 0;
    clear_flg = 0;
    in_count = 1;
    maxcode = MAXCODE (n_bits = g_init_bits);

    ClearCode = (1 << (init_bits - 1));
    EOFCode = ClearCode + 1;
    free_ent = ClearCode + 2;

    char_init ();

    hshift = 0;
    for (fcode = (long) hsize; fcode < 65536L; fcode *= 2L)
	hshift++;
    hshift = 8 - hshift;	/* set hash code range bound */

    hsize_reg = hsize;
    cl_hash ((long) hsize_reg);	/* clear hash table */

    output (ClearCode);

    for (hh = 0; hh < h; )
    {
	short li;

	li = fread (qldat, 1, llen, inp);
	(cvt) (data, qldat, w, 0);
	row = data;
	for (li = 0; li < w; li++)
	{
	    if (first)
	    {
		first = 0;
		ent = *row++;
		li++;
	    }
	    c = *row++;
	    in_count++;

	    fcode = (long) (((long) c << maxbits) + ent);
	    i = (((int) c << hshift) ^ ent);	/* xor hashing */

	    if (HashTabOf (i) == fcode)
	    {
		ent = CodeTabOf (i);
		continue;
	    }

	    else if ((long) HashTabOf (i) < 0)	/* empty slot */
		goto nomatch;

	    disp = hsize_reg - i;	/* secondary hash (after G. Knott) */
	    if (i == 0)
		disp = 1;

	  probe:
	    if ((i -= disp) < 0)
		i += hsize_reg;

	    if (HashTabOf (i) == fcode)
	    {
		ent = CodeTabOf (i);
		continue;
	    }

	    if ((long) HashTabOf (i) >= 0)
		goto probe;

	  nomatch:
	    output (ent);
	    out_count++;
	    ent = c;

	    if (free_ent < maxmaxcode)
	    {
		CodeTabOf (i) = free_ent++;	/* code -> hashtable */
		HashTabOf (i) = fcode;
	    }
	    else
		cl_block ();
	}			/* end for */
	hh++;
	if (verbose)
	{
	    fprintf (stderr, "%d/%d%c", hh, h, verbose);
	    fflush (stderr);
	}
    }				/* end while */

    /* Put out the final code */
    output (ent);
    out_count++;
    output (EOFCode);
    free(data);
    free(qldat);
	
}


/*****************************************************************
 * TAG( output )
 *
 * Output the given code.
 * Inputs:
 *      code:   A n_bits-bit integer.  If == -1, then EOF.  This assumes
 *              that n_bits =< (long)wordsize - 1.
 * Outputs:
 *      Outputs code to the file.
 * Assumptions:
 *      Chars are 8 bits long.
 * Algorithm:
 *      Maintain a BITS character long buffer (so that 8 codes will
 * fit in it exactly).  Use the VAX insv instruction to insert each
 * code in turn.  When the buffer fills up empty it and start over.
 */

static
unsigned long masks[] =
{0x0000, 0x0001, 0x0003, 0x0007, 0x000F,
 0x001F, 0x003F, 0x007F, 0x00FF,
 0x01FF, 0x03FF, 0x07FF, 0x0FFF,
 0x1FFF, 0x3FFF, 0x7FFF, 0xFFFF};

static void output (code)
int code;
{
    cur_accum &= masks[cur_bits];

    if (cur_bits > 0)
	cur_accum |= ((long) code << cur_bits);
    else
	cur_accum = code;

    cur_bits += n_bits;

    while (cur_bits >= 8)
    {
	char_out ((int) (cur_accum & 0xff));
	cur_accum >>= 8;
	cur_bits -= 8;
    }

    /*
     * If the next entry is going to be too big for the code size,
     * then increase it, if possible.
     */

    if (free_ent > maxcode || clear_flg)
    {

	if (clear_flg)
	{
	    maxcode = MAXCODE (n_bits = g_init_bits);
	    clear_flg = 0;
	}
	else
	{
	    n_bits++;
	    if (n_bits == maxbits)
		maxcode = maxmaxcode;
	    else
		maxcode = MAXCODE (n_bits);
	}
    }

    if (code == EOFCode)
    {
	/* At EOF, write the rest of the buffer */
	while (cur_bits > 0)
	{
	    char_out ((int) (cur_accum & 0xff));
	    cur_accum >>= 8;
	    cur_bits -= 8;
	}

	flush_char ();

	fflush (g_outfile);

#ifdef FOO
	if (ferror (g_outfile))
	    FatalError ("unable to write GIF file");
#endif
    }
}


/********************************/
static void cl_block ()		/* table clear for block compress */
{
    /* Clear out the hash table */

    cl_hash ((long) hsize);
    free_ent = ClearCode + 2;
    clear_flg = 1;

    output (ClearCode);
}


/********************************/
static void cl_hash (hsize)	/* reset code table */
long hsize;
{
    long *htab_p = htab + hsize;
    long i;
    long m1 = -1;

    i = hsize - 16;
    do
    {				/* might use Sys V memset(3) here */
	*(htab_p - 16) = m1;
	*(htab_p - 15) = m1;
	*(htab_p - 14) = m1;
	*(htab_p - 13) = m1;
	*(htab_p - 12) = m1;
	*(htab_p - 11) = m1;
	*(htab_p - 10) = m1;
	*(htab_p - 9) = m1;
	*(htab_p - 8) = m1;
	*(htab_p - 7) = m1;
	*(htab_p - 6) = m1;
	*(htab_p - 5) = m1;
	*(htab_p - 4) = m1;
	*(htab_p - 3) = m1;
	*(htab_p - 2) = m1;
	*(htab_p - 1) = m1;
	htab_p -= 16;
    }
    while ((i -= 16) >= 0);

    for (i += 16; i > 0; i--)
	*--htab_p = m1;
}


/******************************************************************************
 *
 * GIF Specific routines
 *
 ******************************************************************************/

/*
 * Number of characters so far in this 'packet'
 */
static int a_count;

/*
 * Set up the 'byte output' routine
 */
static void char_init ()
{
    a_count = 0;
}

/*
 * Define the storage for the packet accumulator
 */
static char accum[256];

/*
 * Add a character to the end of the current packet, and if it is 254
 * characters, flush the packet to disk.
 */
static void char_out (c)
int c;
{
    accum[a_count++] = c;
    if (a_count >= 254)
	flush_char ();
}

/*
 * Flush the packet to disk, and reset the accumulator
 */
static void flush_char ()
{
    if (a_count > 0)
    {
	fputc (a_count, g_outfile);
	fwrite (accum, (size_t) 1, (size_t) a_count, g_outfile);
	a_count = 0;
    }
}
#endif

static void writeBMP1 (FILE * fp, FILE * inp, int w, int h, int llen,
		       void (*func) (uchar *, uchar *, short, short))
{
    int i, padlen;
    uchar *qldat, *buf;

    qldat = xmalloc (llen);
    padlen = ((w + 31) / 32) * 4;	/* needs to be a multiple of 32 bits */
    buf = xmalloc (llen + 8);

    for (i = h - 1; i >= 0; i--)
    {
	fseek (inp, 10 + (i * llen), SEEK_SET);
	fread (qldat, 1, llen, inp);
	(func) (buf, qldat, w, 1);
	fwrite (buf, 1, padlen, fp);
	if (verbose)
	{
	    fprintf (stderr, "%d/%d%c", h - i, h, verbose);
	    fflush (stderr);
	}
    }
    free(qldat);
    free(buf);
}

static void writeBMP4 (FILE * fp, FILE * inp, int w, int h, int llen,
		       void (*func) (uchar *, uchar *, short, short))
{
    int padlen, i, j;
    uchar *qldat, *buf, *p;
    uchar c = 0;

    buf = xmalloc (llen * 8 + 8);
    qldat = xmalloc (llen);

    padlen = ((w + 7) / 8) * 4;	/* padded to a multiple of 32 bits */

    for (i = h - 1; i >= 0; i--)
    {
	fseek (inp, 10 + (i * llen), SEEK_SET);
	fread (qldat, 1, llen, inp);
	(func) (buf, qldat, w, 0);
	p = buf;
	for (j = 0; j < w; j++)
	{
	    if (j & 1)
	    {
		c <<= 4;
		c |= *p++;
		putc (c, fp);
	    }
	    else
	    {
		c = *p++;
	    }
	}
	if (verbose)
	{
	    fprintf (stderr, "%d/%d%c", h - i, h, verbose);
	    fflush (stderr);
	}
    }
    free(qldat);
    free(buf);
}

#ifdef __GNUC__
# define PACKED __attribute__ ((packed))
#else
# define PACKED
#endif

typedef struct
{
    char magno[2] PACKED;
    long fsize PACKED;
    short res1 PACKED;
    short res2 PACKED;
    long offset PACKED;
    int isize PACKED;
    int wid PACKED;
    int hgt PACKED;
    short planes PACKED;
    short nbits PACKED;
    long comp PACKED;
    long rawsize PACKED;
    long xpixm PACKED;
    long ypixm PACKED;
    long ncol PACKED;
    long nicol PACKED;
}
bmp_t;

int WriteBMP (FILE * fp, FILE * inp, int w, int h, int llen,
	      uchar * rmap, uchar * gmap, uchar * bmap, int nc,
	      void (*func) (uchar *, uchar *, short, short))
{
    int i, nbits, blin;
    bmp_t bh;

    nbits = 0;

    if (nc != 2)
    {
	nbits = 4;
    }
    else
	nbits = 1;

    blin = ((w * nbits + 31) / 32) * 4;		/* bytes per line */

    bh.magno[0] = 'B';
    bh.magno[1] = 'M';

    bh.fsize = VAXLONG (sizeof (bmp_t) + (nc * 4) + blin * h);
    bh.res1 = bh.res2 = 0;
    bh.offset = VAXLONG (sizeof (bmp_t) + (nc * 4));
    bh.isize = VAXLONG (sizeof (bmp_t) - offsetof (bmp_t, isize));
    bh.wid = VAXLONG (w);
    bh.hgt = VAXLONG (h);
    bh.planes = VAXSHORT (1);
    bh.nbits = VAXSHORT (nbits);
    bh.comp = 0;
    bh.rawsize = VAXLONG (blin * h);
    bh.xpixm = VAXLONG (75 * 39);
    bh.ypixm = VAXLONG (75 * 39);
    bh.ncol = bh.nicol = VAXLONG (nc);

    fwrite (&bh, sizeof (bh), 1, fp);

    for (i = 0; i < nc; i++)
    {
	putc (bmap[i], fp);
	putc (gmap[i], fp);
	putc (rmap[i], fp);
	putc (0, fp);
    }

    if (nbits == 1)
	writeBMP1 (fp, inp, w, h, llen, func);
    else
	writeBMP4 (fp, inp, w, h, llen, func);
    return 0;
}

/* pnmtops.c - read a portable anymap and produce a PostScript file
   **
   ** Copyright (C) 1989 by Jef Poskanzer.
   **
   ** Permission to use, copy, modify, and distribute this software and its
   ** documentation for any purpose and without fee is hereby granted, provided
   ** that the above copyright notice appear in all copies and that both that
   ** copyright notice and this permission notice appear in supporting
   ** documentation.  This software is provided "as is" without express or
   ** implied warranty.
   **
   **
   ** -nocenter option added November 1993 by Wolfgang Stuerzlinger,
   **  wrzl@gup.uni-linz.ac.at.
   **
 */

/*
 * Somewhat hacked by jh for unpic
 */

#define MARGIN 0.95
#define _GETR(p) (((p) & 1) ? bmask : 0)
#define _GETG(p) (((p) & 2) ? bmask : 0)
#define _GETB(p) (((p) & 4) ? bmask : 0)

static int bitspersample, item, bitsperitem, bitshift, itemsperline, items;
static int rleitem, rlebitsperitem, rlebitshift;
static int repeat, itembuf[128], count, repeatitem, repeatcount;

static void
 putinit (FILE * fp, char *name, int cols, int rows, int padright, int bps,
	  float scale, int dpi, int pagewid, int pagehgt,
	  int rleflag, int center)
{
    int icols, irows, devpix;
    float pixfac, scols, srows, llx, lly;

    /* Turn? */
    icols = cols;
    irows = rows;

    /* Figure out size. */
    devpix = dpi / 72.0 + 0.5;	/* device pixels per unit, approx. */
    pixfac = 72.0 / dpi * devpix;	/* 1, approx. */
    scols = scale * cols * pixfac;
    srows = scale * rows * pixfac;
    if (scols > pagewid * MARGIN || srows > pagehgt * MARGIN)
    {
	if (scols > pagewid * MARGIN)
	{
	    scale *= pagewid / scols * MARGIN;
	    scols = scale * cols * pixfac;
	    srows = scale * rows * pixfac;
	}
	if (srows > pagehgt * MARGIN)
	{
	    scale *= pagehgt / srows * MARGIN;
	    scols = scale * cols * pixfac;
	    srows = scale * rows * pixfac;
	}
    }
    llx = (center) ? (pagewid - scols) / 2 : 0;
    lly = (center) ? (pagehgt - srows) / 2 : 0;

    fprintf (fp, "%%!PS-Adobe-2.0 EPSF-2.0\n");
    fprintf (fp, "%%%%Creator: unpictops\n");
    fprintf (fp, "%%%%Title: %s\n", name);
    fprintf (fp, "%%%%Pages: 1\n");
    fprintf (fp,
	     "%%%%BoundingBox: %d %d %d %d\n",
	     (int) llx, (int) lly,
	     (int) (llx + scols + 0.5), (int) (lly + srows + 0.5));
    fprintf (fp, "%%%%EndComments\n");
    if (rleflag)
    {
	fprintf (fp, "/rlestr1 1 string def\n");
	fprintf (fp, "/readrlestring {\n");	/* s -- nr */
	fprintf (fp, "  /rlestr exch def\n");	/* - */
	fprintf (fp, "  currentfile rlestr1 readhexstring pop\n");	/* s1 */
	fprintf (fp, "  0 get\n");	/* c */
	fprintf (fp, "  dup 127 le {\n");	/* c */
	fprintf (fp, "    currentfile rlestr 0\n");	/* c f s 0 */
	fprintf (fp, "    4 3 roll\n");		/* f s 0 c */
	fprintf (fp, "    1 add  getinterval\n");	/* f s */
	fprintf (fp, "    readhexstring pop\n");	/* s */
	fprintf (fp, "    length\n");	/* nr */
	fprintf (fp, "  } {\n");	/* c */
	fprintf (fp, "    256 exch sub dup\n");		/* n n */
	fprintf (fp, "    currentfile rlestr1 readhexstring pop\n");	/* n n s1 */
	fprintf (fp, "    0 get\n");	/* n n c */
	fprintf (fp, "    exch 0 exch 1 exch 1 sub {\n");	/* n c 0 1 n-1 */
	fprintf (fp, "      rlestr exch 2 index put\n");
	fprintf (fp, "    } for\n");	/* n c */
	fprintf (fp, "    pop\n");	/* nr */
	fprintf (fp, "  } ifelse\n");	/* nr */
	fprintf (fp, "} bind def\n");
	fprintf (fp, "/readstring {\n");	/* s -- s */
	fprintf (fp, "  dup length 0 {\n");	/* s l 0 */
	fprintf (fp, "    3 copy exch\n");	/* s l n s n l */
	fprintf (fp, "    1 index sub\n");	/* s l n s n r */
	fprintf (fp, "    getinterval\n");	/* s l n ss */
	fprintf (fp, "    readrlestring\n");	/* s l n nr */
	fprintf (fp, "    add\n");	/* s l n */
	fprintf (fp, "    2 copy le { exit } if\n");	/* s l n */
	fprintf (fp, "  } loop\n");	/* s l l */
	fprintf (fp, "  pop pop\n");	/* s */
	fprintf (fp, "} bind def\n");
    }
    else
    {
	fprintf (fp, "/readstring {\n");	/* s -- s */
	fprintf (fp, "  currentfile exch readhexstring pop\n");
	fprintf (fp, "} bind def\n");
    }
    if (bps > 1)
    {
	fprintf (fp, "/rpicstr %d string def\n", (icols + padright) * bps / 8);
	fprintf (fp, "/gpicstr %d string def\n", (icols + padright) * bps / 8);
	fprintf (fp, "/bpicstr %d string def\n", (icols + padright) * bps / 8);
    }
    else
	fprintf (fp, "/picstr %d string def\n", (icols + padright) * bps / 8);
    fprintf (fp, "%%%%EndProlog\n");
    fprintf (fp, "%%%%Page: 1 1\n");
    fprintf (fp, "gsave\n");
    fprintf (fp, "%g %g translate\n", llx, lly);
    fprintf (fp, "%g %g scale\n", scols, srows);
    fprintf (fp, "%d %d %d\n", icols, irows, bps);
    fprintf (fp, "[ %d 0 0 -%d 0 %d ]\n", icols, irows, irows);
    if (bps > 1)
    {
	fprintf (fp, "{ rpicstr readstring }\n");
	fprintf (fp, "{ gpicstr readstring }\n");
	fprintf (fp, "{ bpicstr readstring }\n");
	fprintf (fp, "true 3\n");
	fprintf (fp, "colorimage\n");
    }
    else
    {
	fprintf (fp, "{ picstr readstring }\n");
	fprintf (fp, "image\n");
    }

    bitspersample = bps;
    itemsperline = items = 0;
    if (rleflag)
    {
	rleitem = 0;
	rlebitsperitem = 0;
	rlebitshift = 8 - bitspersample;
	repeat = 1;
	count = 0;
    }
    else
    {
	item = 0;
	bitsperitem = 0;
	bitshift = 8 - bitspersample;
    }
}

static void
 putitem (FILE * fp)
{
    char *hexits = "0123456789abcdef";

    if (itemsperline == 30)
    {
	fputc ('\n', fp);
	itemsperline = 0;
    }
    fputc (hexits[item >> 4], fp);
    fputc (hexits[item & 15], fp);
    ++itemsperline;
    ++items;
    item = 0;
    bitsperitem = 0;
    bitshift = 8 - bitspersample;
}

static void putxelval (FILE * fp, uchar xv)
{
    if (bitsperitem == 8)
	putitem (fp);
    item += xv << bitshift;
    bitsperitem += bitspersample;
    bitshift -= bitspersample;
}

static void
 putrest (FILE * fp)
{
    if (bitsperitem > 0)
	putitem (fp);
    fprintf (fp, "\n");
    fprintf (fp, "grestore\n");
    fprintf (fp, "showpage\n");
    fprintf (fp, "%%%%Trailer\n");
}

static void
 rleputbuffer (FILE * fp)
{
    int i;

    if (repeat)
    {
	item = 256 - count;
	putitem (fp);
	item = repeatitem;
	putitem (fp);
    }
    else
    {
	item = count - 1;
	putitem (fp);
	for (i = 0; i < count; ++i)
	{
	    item = itembuf[i];
	    putitem (fp);
	}
    }
    repeat = 1;
    count = 0;
}

static void
 rleputitem (FILE * fp)
{
    int i;

    if (count == 128)
	rleputbuffer (fp);

    if (repeat && count == 0)
    {				/* Still initializing a repeat buf. */
	itembuf[count] = repeatitem = rleitem;
	++count;
    }
    else if (repeat)
    {				/* Repeating - watch for end of run. */
	if (rleitem == repeatitem)
	{			/* Run continues. */
	    itembuf[count] = rleitem;
	    ++count;
	}
	else
	{			/* Run ended - is it long enough to dump? */
	    if (count > 2)
	    {			/* Yes, dump a repeat-mode buffer and start a new one. */
		rleputbuffer (fp);
		itembuf[count] = repeatitem = rleitem;
		++count;
	    }
	    else
	    {			/* Not long enough - convert to non-repeat mode. */
		repeat = 0;
		itembuf[count] = repeatitem = rleitem;
		++count;
		repeatcount = 1;
	    }
	}
    }
    else
    {				/* Not repeating - watch for a run worth repeating. */
	if (rleitem == repeatitem)
	{			/* Possible run continues. */
	    ++repeatcount;
	    if (repeatcount > 3)
	    {			/* Long enough - dump non-repeat part and start repeat. */
		count = count - (repeatcount - 1);
		rleputbuffer (fp);
		count = repeatcount;
		for (i = 0; i < count; ++i)
		    itembuf[i] = rleitem;
	    }
	    else
	    {			/* Not long enough yet - continue as non-repeat buf. */
		itembuf[count] = rleitem;
		++count;
	    }
	}
	else
	{			/* Broken run. */
	    itembuf[count] = repeatitem = rleitem;
	    ++count;
	    repeatcount = 1;
	}
    }

    rleitem = 0;
    rlebitsperitem = 0;
    rlebitshift = 8 - bitspersample;
}

static void rleputxelval (FILE * fp, uchar xv)
{
    if (rlebitsperitem == 8)
	rleputitem (fp);
    rleitem += xv << rlebitshift;
    rlebitsperitem += bitspersample;
    rlebitshift -= bitspersample;
}

static void
 rleflush (FILE * fp)
{
    if (rlebitsperitem > 0)
	rleputitem (fp);
    if (count > 0)
	rleputbuffer (fp);
}

static void
 rleputrest (FILE * fp)
{
    rleflush (fp);
    fprintf (fp, "\n");
    fprintf (fp, "grestore\n");
    fprintf (fp, "showpage\n");
    fprintf (fp, "%%%%Trailer\n");
}

int WritePS (FILE * fp, FILE * ifp, int w, int h, int llen,
	     uchar * rmap, uchar * gmap, uchar * bmap,
	     int nc, void (*func) (uchar *, uchar *, short, short))
{
    uchar *qldat, *buf;
    uchar *xP;
    int argn, rleflag, center;
    int rows, cols, bps = 0, padright, row, col;
    float scale;
    int dpi, pagewid, pagehgt;
    uchar bmask = 0;

    argn = 1;
    scale = 1.0;
    center = 1;
    rleflag = 1;
    /* LaserWriter defaults. */
    dpi = 300;
    pagewid = 612;
    pagehgt = 762;

    switch (nc)
    {
	case 2:
	    bps = 1;
	    bmask = 1;
	    break;
	case 4:
	    bps = 2;
	    bmask = 3;
	    break;
	case 8:
	    bps = 4;
	    bmask = 15;
	    break;
    }

    rows = h;
    cols = w;

    padright = (((cols * bps + 7) / 8) * 8 - cols * bps) / bps;

    putinit (fp, "unpic2ps", cols, rows, padright, bps, scale, dpi,
	     pagewid, pagehgt, rleflag, center);

    buf = xmalloc (llen * 8 + 8);
    qldat = xmalloc (llen);

    for (row = 0; row < rows; )
    {
	fread (qldat, 1, llen, ifp);
	(func) (buf, qldat, w, 0);

	if (nc == 2)
	{
	    for (col = 0, xP = buf; col < cols; ++col, ++xP)
	    {
		if (rleflag)
		    rleputxelval (fp, (*xP));
		else
		    putxelval (fp, (*xP));
	    }
	    for (col = 0; col < padright; ++col)
		if (rleflag)
		    rleputxelval (fp, 0);
		else
		    putxelval (fp, 0);
	    if (rleflag)
		rleflush (fp);
	}
	else
	{
	    /* First red. */
	    for (col = 0, xP = buf; col < cols; ++col, ++xP)
	    {
		if (rleflag)
		    rleputxelval (fp, _GETR (*xP));
		else
		    putxelval (fp, _GETR (*xP));
	    }
	    for (col = 0; col < padright; ++col)
		if (rleflag)
		    rleputxelval (fp, 0);
		else
		    putxelval (fp, 0);
	    if (rleflag)
		rleflush (fp);
	    /* Then green. */
	    for (col = 0, xP = buf; col < cols; ++col, ++xP)
	    {
		if (rleflag)
		    rleputxelval (fp, _GETG (*xP));
		else
		    putxelval (fp, _GETG (*xP));
	    }
	    for (col = 0; col < padright; ++col)
	    {
		if (rleflag)
		    rleputxelval (fp, 0);
		else
		    putxelval (fp, 0);
	    }
	    if (rleflag)
		rleflush (fp);
	    /* And blue. */
	    for (col = 0, xP = buf; col < cols; ++col, ++xP)
	    {
		uchar q;

		if (nc == 8)
		{
		    q = _GETB (*xP);
		}
		else
		{
		    q = (*xP == 3) ? bmask : 0;
		}
		if (rleflag)
		    rleputxelval (fp, q);
		else
		    putxelval (fp, q);
	    }
	    for (col = 0; col < padright; ++col)
		if (rleflag)
		    rleputxelval (fp, 0);
		else
		    putxelval (fp, 0);
	    if (rleflag)
		rleflush (fp);
	}

	row++;
	if (verbose)
	{
	    fprintf (stderr, "%d/%d%c", row, rows, verbose);
	    fflush (stderr);
	}
    }

    free(qldat);
    free(buf);

    if (rleflag)
	rleputrest (fp);
    else
	putrest (fp);

    return 0;
}


/* ppmtopcx.c - read a portable pixmap and produce a PCX file
   **
   ** Copyright (C) 1990 by Michael Davidson.
   **
   ** Permission to use, copy, modify, and distribute this software and its
   ** documentation for any purpose and without fee is hereby granted, provided
   ** that the above copyright notice appear in all copies and that both that
   ** copyright notice and this permission notice appear in supporting
   ** documentation.  This software is provided "as is" without express or
   ** implied warranty.
 */

static void putplane (FILE * fp, unsigned char *buf, int size)
{
    uchar *end;
    int c;
    int previous;
    int count;

    end = buf + size;

    previous = *buf++;
    count = 1;

    while (buf < end)
    {
	c = *buf++;
	if (c == previous && count < 63)
	{
	    ++count;
	    continue;
	}

	if (count > 1 || (previous & 0xc0) == 0xc0)
	{
	    count |= 0xc0;
	    fputc (count, fp);
	}
	fputc (previous, fp);
	previous = c;
	count = 1;
    }

    if (count > 1 || (previous & 0xc0) == 0xc0)
    {
	count |= 0xc0;
	fputc (count, fp);
    }
    fputc (previous, fp);
}


int WritePCX (FILE * fp, FILE * ifp, int w, int h, int llen,
	      uchar * rmap, uchar * gmap, uchar * bmap,
	      int nc, char *comment,
	      void (*func) (uchar *, uchar *, short, short))
{
    pcx_header ph;
    uchar *qldat, *buf, *ptr, *q;
    short l;
    short i;
    short j;
    short k, m;

    ph.bpp = 1;
    if (nc == 2)
    {
	ph.nplanes = 1;
    }
    else if (nc == 4)
    {
	ph.nplanes = 2;
    }
    else
    {
	ph.nplanes = 3;
    }

    for (i = j = 0; i < nc; i++)
    {
	ph.palette[j] = rmap[i];
	j++;
	ph.palette[j] = gmap[i];
	j++;
	ph.palette[j] = bmap[i];
	j++;
    }

    for (; i < 16; i++)
    {
	ph.palette[j] = 255;
	j++;
	ph.palette[j] = 255;
	j++;
	ph.palette[j] = 255;
	j++;
    }


    buf = xmalloc (llen * 8 + 8);
    qldat = xmalloc (llen);
    l = (w + 7) / 8;
    l = (l + 1) & ~1;
    ph.manuf = 10;
    ph.version = 5;
    ph.encoding = 1;
    ph.x1 = 0;
    ph.y1 = 0;
    ph.x2 = VAXSHORT (w - 1);
    ph.y2 = VAXSHORT (h - 1);
    ph.hres = w;
    ph.vres = h;
    ph.reserved = 0;
    ph.bpl = VAXSHORT (l);
    ph.palinfo = 1;

    /* Write the header. */

    fwrite (&ph, 1, 128, fp);

    for (i = 0; i < h;)
    {
	fread (qldat, 1, llen, ifp);
	(func) (buf, qldat, w, (nc == 2));

	for (j = 0; j < ph.nplanes; j++)
	{
	    if (nc == 2)
	    {
		ptr = buf;
	    }
	    else
	    {
		q = ptr = qldat;
		memset (qldat, 0, llen);
		for (k = m = 0; k < w; k++)
		{
		    if (*(buf + k) & (1 << j))
		    {
			*q |= (1 << (7 - m));
		    }
		    m++;
		    m &= 7;
		    if (m == 0)
			q++;
		}
	    }
	    putplane (fp, ptr, l);
	}

	i++;
	if (verbose)
	{
	    fprintf (stderr, "%d/%d%c", i, h, verbose);
	    fflush (stderr);
	}
    }
    free (buf);
    free (qldat);
    return 0;
}


void WritePNG (FILE * fp, FILE * inp, int w, int h, int llen,
	       uchar * rmap, uchar * gmap, uchar * bmap, int nc,
	       void (*func) (uchar *, uchar *, short, short))
{
#ifdef __GNUC__
    extern char png_libpng_ver[] __attribute__ ((unused));// Just to Shutup GCC
#endif
#include "png.h"
#define AUTHOR "Jonathan R Hudson"
#define SOFTWARE "unpic for " OSNAME
#define COPYRIGHT "(c) 1997 All rights reserved"

    short i, j;
    uchar *qldat, *buf;

    png_infop info_ptr;
    png_colorp cmap;
    png_text txt[3];

    png_structp png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
						   NULL, NULL, NULL);

    if (!png_ptr)
	return;

    info_ptr = png_create_info_struct (png_ptr);
    if (!info_ptr)
    {
	png_destroy_write_struct (&png_ptr, (png_infopp) NULL);
	return;
    }

    if (setjmp (png_ptr->jmpbuf))
    {
	png_destroy_write_struct (&png_ptr, &info_ptr);
	return;
    }
    cmap = xmalloc (nc * sizeof (png_color));

    for (j = 0; j < nc; j++)
    {
	(cmap + j)->red = rmap[j];
	(cmap + j)->green = gmap[j];
	(cmap + j)->blue = bmap[j];
    }

    txt[0].compression = -1;
    txt[0].key = "Author";
    txt[0].text = AUTHOR;
    txt[0].text_length = sizeof (AUTHOR) - 1;

    txt[1].compression = -1;
    txt[1].key = "Copyright";
    txt[1].text = COPYRIGHT;
    txt[1].text_length = sizeof (COPYRIGHT) - 1;

    txt[2].compression = -1;
    txt[2].key = "Software";
    txt[2].text = SOFTWARE;
    txt[2].text_length = sizeof (SOFTWARE) - 1;

    png_write_init (png_ptr);

    png_init_io (png_ptr, fp);

    info_ptr->width = w;
    info_ptr->height = h;
    info_ptr->num_palette = nc;

    switch (nc)
    {
	case 2:
	    info_ptr->rowbytes = (w + 7) / 8;
	    info_ptr->bit_depth = 1;
	    break;
	case 4:
	    info_ptr->rowbytes = (w + 3) / 4;
	    info_ptr->bit_depth = 2;
	    break;
	case 8:
	    info_ptr->rowbytes = (w + 1) / 2;
	    info_ptr->bit_depth = 4;
	    break;
    }

    info_ptr->palette = cmap;
    info_ptr->num_trans = 0;
    info_ptr->color_type = PNG_COLOR_TYPE_PALETTE;
    info_ptr->compression_type = 0;
    info_ptr->filter_type = 0;
    info_ptr->interlace_type = 0;
    info_ptr->text = txt;
    info_ptr->num_text = 3;
    info_ptr->valid |= PNG_INFO_PLTE;

    png_write_info (png_ptr, info_ptr);

    qldat = xmalloc (llen);
    buf = xmalloc (llen * 8 + 8);

    for (i = 0; i < h; )
    {
	fread (qldat, 1, llen, inp);
	(func) (buf, qldat, w, info_ptr->bit_depth);
	png_write_row (png_ptr, buf);
	i++;
	if (verbose)
	{
	    fprintf (stderr, "%d/%d%c", i, h, verbose);
	    fflush (stderr);
	}
    }
    png_write_end (png_ptr, NULL);
    png_destroy_write_struct (&png_ptr, &info_ptr);

    free(qldat);
    free(buf);
    free (cmap);
}

#ifndef O_BINARY
# define O_BINARY 0
#endif

int main (int ac, char **av)
{
    QLS qs;
    struct stat s;
    short cols = 0, typ[2];
    void (*func) (uchar *, uchar *, short, short) = NULL;
    short defmode = 0;
    int c;
    short forked = 0;

    char *ip = NULL;
    char *op = NULL;

    qs.qlar = 4;
    qs.flag = 0;
    typ[1] = 0;

    while (
#ifdef LZW_CLEAN
	      (c = getopt (ac, av, "#!VxXbBpPvsSm:"))
#else
	      (c = getopt (ac, av, "#!VxXbBgtpGT:PsSvm:"))
#endif
	      != EOF)
    {

	switch (c)
	{
	    case 'T':
		typ[1] = *optarg;
	    case 't':
	    case 'G':
	    case 'P':
	    case 'g':
	    case 'p':
	    case 'B':
	    case 'b':
	    case 'S':
	    case 's':
	    case 'X':
	    case 'x':
		typ[0] = c;
		break;
	    case 'm':
		defmode = strtol (optarg, 0, 0);
		break;
	    case 'v':
		verbose = '\r';
		break;
	    case 'V':
		fputs ("unpic v" VERSION " for " OSNAME "\n", stderr);
		exit (0);
		break;
	    case '#':
		setvbuf(stdout, NULL, _IONBF, 0);
		setvbuf(stdin, NULL, _IONBF, 0);
		write(1, METHODS, sizeof(METHODS));
		forked = 1;
#ifdef QDOS
		{
		    extern char *_endmsg;
		    _endmsg = NULL;
		}
#endif
		break;
	    default:
		usage ();
		break;
	}
    }

    for (; !forked && optind < ac; optind++)
    {
	if (ip == NULL)
	{
	    ip = *(av + optind);
	}
	else
	{
	    op = *(av + optind);
	}
    }

    if (ip)
    {
	qs.inp = fopen (ip, "rb");
    }
    else
    {
	qs.inp = stdin;
    }

    if (qs.inp)
    {
	do 
	{
	    if(forked)
	    {
		char cmd[PATH_MAX];
		if(fgets(cmd, sizeof(cmd), qs.inp))
		{
		    char *nl;
		    typ[0] = *cmd;
		    typ[1] = *(cmd + 1);
		    op = cmd + 2;
		    if((nl = strchr(op, '\n')))
		    {
			*nl = 0;
		    }
		}
		else
		{
		    forked = 0;
		    break;
		}
	    }
	    if (fread (&qs.qh, 1, sizeof (QHEAD), qs.inp) > 0)
	    {
#ifdef VAX_ENDIAN
		qs.qh.flag = swapword (qs.qh.flag);
		qs.qh.xpix = swapword (qs.qh.xpix);
		qs.qh.ypix = swapword (qs.qh.ypix);
		qs.qh.llen = swapword (qs.qh.llen);
#endif

		if (qs.qh.flag != 0x4afc)
		{
		    fstat (fileno (qs.inp), &s);
		    qs.qh.xpix = 512;
		    qs.qh.llen = 128;
		    qs.qh.ypix = (s.st_size + 127) / 128;
		    qs.qh.mode[0] = defmode;
		    fseek (qs.inp, 0, 0);
		}

		if (op)
		{
		    qs.fp = fopen (op, "wb");
		}
		else
		{
		    qs.fp = stdout;
		}
		
		if (qs.fp)
		{
		    uchar *r = NULL, *g = NULL, *b = NULL;
		    
		    uchar bw[] =
		    {0, 0xff};
		    
		    uchar r4map[] =
		    {0, 0xff, 0, 0xff};
		    uchar g4map[] =
		    {0, 0, 0xff, 0xff};
		    uchar b4map[] =
		    {0, 0, 0, 0xff};
		    
/*                            0    1     2    3     4     5      6    7 */
/*                           blk red green  yellw  blue mgent  cyan  white */
		    uchar r8map[] =
		    {0, 0xff, 0, 0xff, 0, 0xff, 0, 0xff};
		    uchar g8map[] =
		    {0, 0, 0xff, 0xff, 0, 0, 0xff, 0xff};
		    uchar b8map[] =
		    {0, 0, 0, 0, 0xff, 0xff, 0xff, 0xff};
		    
		    if (qs.qh.mode[0] == 0)
		    {
			cols = 4;
			func = MakeChunkyPix4;
			r = r4map;
			g = g4map;
			b = b4map;
		    }
		    else if (qs.qh.mode[0] == 8)
		    {
			cols = 8;
			func = MakeChunkyPix8;
			r = r8map;
			g = g8map;
			b = b8map;
		    }
		    else if (qs.qh.mode[0] == 2)
		    {
			cols = 2;
			r = g = b = bw;
			func = MakeChunkyPix2;
		    }
		    else
		    {
			exit (0);
		    }
		    
		    fprintf (stderr, "Size %d x %d (%d), %d\n",
			     qs.qh.xpix, qs.qh.ypix, qs.qh.llen, cols);

		    if (tolower (typ[0]) == 'x')
		    {
			WritePCX (qs.fp, qs.inp,
				  qs.qh.xpix, qs.qh.ypix, qs.qh.llen,
				  r, g, b, cols, 
				  "unpic (c) Jonathan Hudson 1997",
				  func);
			fclose (qs.fp);
		    }
		    else if (tolower (typ[0]) == 'b')
		    {
			WriteBMP (qs.fp, qs.inp,
				  qs.qh.xpix, qs.qh.ypix, qs.qh.llen,
				  r, g, b, cols, func);
			
			fclose (qs.fp);
		    }
		    else if (tolower (typ[0]) == 'p')
		    {
			WritePNG (qs.fp, qs.inp,
				  qs.qh.xpix, qs.qh.ypix, qs.qh.llen,
				  r, g, b, cols, func);
			
			fclose (qs.fp);
		    }
		    else if (tolower (typ[0]) == 's')
		    {
			WritePS (qs.fp, qs.inp,
				 qs.qh.xpix, qs.qh.ypix, qs.qh.llen,
				 r, g, b, cols, func);
			fclose (qs.fp);
		    }
#ifndef LZW_CLEAN
		    else if (tolower (typ[0]) == 'g')
		    {
			ObnoxiousMsg ();
			WriteGIF (qs.fp, qs.inp,
				  qs.qh.xpix, qs.qh.ypix, qs.qh.llen,
				  r, g, b, cols, 
				  "unpic (c) Jonathan Hudson 1997",
				  func);
			fclose (qs.fp);
		    }
#ifndef NOTIF
		    else if (tolower (typ[0]) == 't')
		    {
			short comp;
			comp = COMPRESSION_LZW;
			if (tolower (typ[1]) == 'p')
			    comp = (short) COMPRESSION_PACKBITS;
			else if (tolower (typ[1]) == 'n')
			    comp = COMPRESSION_NONE;
			
			fclose (qs.fp);
			WriteTIFF (op, qs.inp,
				   qs.qh.xpix, qs.qh.ypix, qs.qh.llen,
				   r, g, b, cols, comp,
				   "unpic (c) Jonathan Hudson 1997",
				   func);
		    }
#else
		    else
			fputs ("No Tiff on Win/DOS\n", stderr);
#endif
#endif
		}
		else
		{
		    perror ("output");
		}
		if (verbose)
		    fputc ('\n', stderr);
	    }
	} while (forked);
    }
    else
    {
	perror ("input");
    }
    return 0;
}
